<?php

/**
 * @file
 * Contains \Drupal\cdek_api\Cdek.
 */

namespace Drupal\cdek_api;

use Drupal\cdek_api\Exception\CalculatorApiException;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Symfony\Component\Serializer\Serializer;
use Drupal\Component\Utility\Unicode;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Cache\Cache;

/**
 * Wrapper class for CDEK API.
 *
 * See CDEK API documentation for more info.
 */
class Cdek {

  /**
   * The key for the "all" parameter.
   */
  const PARAMS_ALL = 'all';

  /**
   * CDEK API settings.
   *
   * @var array
   */
  protected $settings = [];

  /**
   * CDEK API request manager.
   *
   * @var \Drupal\cdek_api\RequestManager
   */
  protected $requestManager;

  /**
   * The cache object associated with the 'cdek_api' bin.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The serializer.
   *
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected $serializer;

  /**
   * Cdek constructor.
   *
   * @param \Drupal\cdek_api\RequestManager $request_manager
   *   CDEK API request manager.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache object associated with the 'cdek_api' bin.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   */
  public function __construct(RequestManager $request_manager, CacheBackendInterface $cache, ModuleHandlerInterface $module_handler) {
    $this->requestManager = $request_manager;
    $this->cache = $cache;
    $this->moduleHandler = $module_handler;
    // Create an instance of the serializer.
    $encoders[] = new JsonEncoder();
    $this->serializer = new Serializer([], $encoders);
    // Get the configuration values.
    $this->settings = self::getSettings();
  }

  /**
   * Gets CDEK API settings.
   *
   * @return array
   *   Array with settings.
   */
  public static function getSettings() {
    $settings = \Drupal::config('cdek_api.settings')->get();
    unset($settings['_core']);
    return $settings;
  }

  /**
   * Load order pickup points.
   *
   * @param array $params
   *   API function params for filtering results.
   *   The following parameters are allowed:
   *   - cityid - City code in CDEK system.
   *   - citypostcode - City postcode string, cityid is more priority.
   *   - type - Pickup point type. Allowed values: "PVZ", "POSTOMAT", "ALL".
   *     Default value is "PVZ".
   *   - countryid - Country code in CDEK system.
   *   - regionid - Region code in CDEK system.
   *   - havecashless - (boolean) Is there a payment terminal.
   *   - isdressingroom - (boolean) Is there a fitting room in the pickup point.
   *   - weightmax - The maximum weight that can be sent to the pickup point.
   *   - allowedcod - (boolean) CASH is allowed in the pickup point.
   *   - lang - Localization of the pickup point. The default is "rus".
   *   - takeonly - (boolean) The pickup point is for taking only.
   *
   * @return array|null
   *   Order pickup points array or NULL on failure.
   */
  public function getOrderPickupPoints(array $params = []) {
    static $pickup_points = [];
    $hash = $this->generateParamsHash($params);

    // Check static data.
    if (isset($pickup_points[$hash])) {
      return $pickup_points[$hash];
    }
    // Check cache data.
    $cid = "order_pickup_points:$hash";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $pickup_points[$hash] = $cache_data;
    }

    // Sending API request.
    $result = $this->requestManager->sendOrderPickupPointsRequest($params);
    if ($result === NULL) {
      return NULL;
    }
    try {
      // Convert JSON to an associative array.
      $data = $this->serializer->decode($result, 'json');
      // Check the presence of the element with order pickup points.
      if (!isset($data['pvz']) || !is_array($data['pvz'])) {
        throw new \UnexpectedValueException('Missing element with order pickup points.');
      }
    }
    catch (\UnexpectedValueException $e) {
      // Failed to decode JSON.
      return NULL;
    }
    $data = $data['pvz'];
    // Use the pickup point code as the key.
    foreach ($data as $key => $value) {
      $data[$value['code']] = $value;
      unset($data[$key]);
    }

    // Sort points alphabetically.
    uasort($data, [$this, 'comparePointsByName']);
    // Trigger hook_cdek_api_order_pickup_points_alter().
    // Allow modules to override the list of order pickup points.
    $this->moduleHandler->alter('cdek_api_order_pickup_points', $data, $params);

    // Set persistent and static cache.
    $pickup_points[$hash] = $data;
    if ($this->isCached()) {
      $this->setCacheData($cid, $pickup_points[$hash]);
    }
    return $pickup_points[$hash];
  }

  /**
   * Load order pickup point.
   *
   * @param string $code
   *   The code of the point to load.
   *
   * @return array|null
   *   The point array or NULL if there is no point with the given code.
   */
  public function getOrderPickupPoint($code) {
    $pickup_points = $this->getOrderPickupPoints();
    return isset($pickup_points[$code]) ? $pickup_points[$code] : NULL;
  }

  /**
   * Load countries.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::getOrderPickupPoints() for details.
   *
   * @return array|null
   *   Country names array keyed by country code, or NULL on failure.
   */
  public function getCountries(array $params = []) {
    static $countries = [];
    $hash = $this->generateParamsHash($params);

    // Check static data.
    if (isset($countries[$hash])) {
      return $countries[$hash];
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get countries from pickup points.
    $countries[$hash] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($countries[$hash][$pickup_point['countryCode']])) {
        $countries[$hash][$pickup_point['countryCode']] = $pickup_point['countryName'];
      }
    }
    // Sort countries alphabetically.
    uasort($countries[$hash], ['Drupal\Component\Utility\Unicode', 'strcasecmp']);
    return $countries[$hash];
  }

  /**
   * Load regions.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::getOrderPickupPoints() for details.
   *
   * @return array|null
   *   Region names array keyed by region code, or NULL on failure.
   */
  public function getRegions(array $params = []) {
    static $regions = [];
    $hash = $this->generateParamsHash($params);

    // Check static data.
    if (isset($regions[$hash])) {
      return $regions[$hash];
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get regions from pickup points.
    $regions[$hash] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($regions[$hash][$pickup_point['regionCode']])) {
        $regions[$hash][$pickup_point['regionCode']] = $pickup_point['regionName'];
      }
    }
    // Sort regions alphabetically.
    uasort($regions[$hash], ['Drupal\Component\Utility\Unicode', 'strcasecmp']);
    return $regions[$hash];
  }

  /**
   * Load cities.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::getOrderPickupPoints() for details.
   *
   * @return array|null
   *   City names array keyed by city code, or NULL on failure.
   */
  public function getCities(array $params = []) {
    static $cities = [];
    $hash = $this->generateParamsHash($params);

    // Check static data.
    if (isset($cities[$hash])) {
      return $cities[$hash];
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get cities from pickup points.
    $cities[$hash] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($cities[$hash][$pickup_point['cityCode']])) {
        $cities[$hash][$pickup_point['cityCode']] = $pickup_point['city'];
      }
    }
    // Sort cities alphabetically.
    uasort($cities[$hash], ['Drupal\Component\Utility\Unicode', 'strcasecmp']);
    return $cities[$hash];
  }

  /**
   * Calculate delivery rate by params.
   *
   * @param array $params
   *   API function params to calculate.
   *   The following parameters are allowed:
   *   - authLogin - Login for CDEK account. Set TRUE to use the login
   *     specified in the configuration.
   *   - secure - Password for CDEK account. If authLogin is set to TRUE,
   *     then the password will be taken from the configuration.
   *   - dateExecute - Planning date of shipment in the format YYYY-ММ-DD.
   *   - senderCityId - Origin city code in CDEK system.
   *   - senderCityPostCode - Zip code of the city of origin.
   *   - receiverCityId - Destination city code in CDEK system.
   *   - receiverCityPostCode - Zip code of the destination city.
   *   - tariffId - Code of chosen tariff.
   *   - modeId - Chosen delivery mode.
   *   - goods - Array with elements each of which contains an
   *     associative array with keys:
   *       - weight - float, piece weight, kg.
   *       - length - integer, piece length, sm.
   *       - width - integer, piece width, sm.
   *       - height - integer, depth, sm.
   *     Or with such kind of parameters:
   *       - weight - float, piece weight, kg.
   *       - volume - float, piece volume, cum.
   *
   * @return array|null
   *   An array with delivery rate information, or NULL on failure.
   *
   * @throws \Drupal\cdek_api\Exception\CalculatorApiException
   *   Delivery rate calculation API exception.
   */
  public function calculateDeliveryRate(array $params = []) {
    // TODO: Define a constant for "version" or move it to a configuration.
    // TODO: Add "tariffList" and "services" params support.
    // Merge to defaults.
    $params += [
      'version' => '1.0',
      'senderCityId' => $this->settings['sender_city_id'],
      'senderCityPostCode' => $this->settings['sender_city_postcode'],
      'tariffId' => $this->settings['tariff_id'],
    ];

    // Use authorization if necessary.
    if (isset($params['authLogin']) && $params['authLogin'] === TRUE) {
      $params['authLogin'] = $this->settings['auth_login'];
      $params['secure'] = $this->settings['secure_password'];
    }
    // Process "secure" param.
    if (isset($params['secure'])) {
      if (!isset($params['dateExecute'])) {
        $params['dateExecute'] = date('Y-m-d', time());
      }
      $params['secure'] = md5($params['dateExecute'] . '&' . $params['secure']);
    }
    // Filter empty values.
    $params = array_filter($params);

    // Sending API request.
    $result = $this->requestManager->sendDeliveryRateCalculationRequest($params);
    if ($result === NULL) {
      return NULL;
    }
    try {
      // Convert JSON to an associative array.
      $data = $this->serializer->decode($result, 'json');
      // Check API errors.
      if (isset($data['error'])) {
        // Throw exception by first error.
        $error = array_shift($data['error']);
        if (isset($error['text']) && isset($error['code'])) {
          throw new CalculatorApiException($error['text'], $error['code']);
        }
      }
      // Check the presence of the result element.
      if (!isset($data['result']) || !is_array($data['result'])) {
        throw new \UnexpectedValueException('Missing result element.');
      }
    }
    catch (\UnexpectedValueException $e) {
      // Failed to decode JSON.
      return NULL;
    }
    return $data['result'];
  }

  /**
   * Verifying that the data is cached.
   *
   * @return bool
   *   API data is cached (TRUE) or not (FALSE).
   */
  public function isCached() {
    return $this->settings['cache_expire'] !== NULL;
  }

  /**
   * Stores API data in the persistent cache.
   *
   * @param string $cid
   *   The cache ID of the data to store.
   * @param mixed $data
   *   The data to store in the cache.
   */
  protected function setCacheData($cid, $data) {
    $expire = $this->settings['cache_expire'];
    if ($expire !== Cache::PERMANENT) {
      $expire += time();
    }
    $this->cache->set($cid, $data, $expire);
  }

  /**
   * Returns API data from the persistent cache.
   *
   * @param string $cid
   *   The cache ID of the data to retrieve.
   *
   * @return mixed|false
   *   The cache data or FALSE on failure.
   */
  protected function getCacheData($cid) {
    $cache = $this->cache->get($cid);
    return $cache && isset($cache->data) ? $cache->data : FALSE;
  }

  /**
   * Calculates hash for params set.
   *
   * @param array $params
   *   A set of params to calculate a hash.
   *
   * @return string
   *   Params set hash.
   */
  protected function generateParamsHash(array $params) {
    if (empty($params)) {
      return self::PARAMS_ALL;
    }
    ksort($params);
    $params = serialize($params);
    return Crypt::hashBase64($params);
  }

  /**
   * Compares two order pickup points by their names.
   *
   * @param array $point1
   *   The first order pickup point.
   * @param array $point2
   *   The second order pickup point.
   *
   * @return int
   *   Returns < 0 if $point1 is less than $point2; > 0 if $point1 is greater
   *   than $point2, and 0 if they are equal.
   */
  protected function comparePointsByName(array $point1, array $point2) {
    return Unicode::strcasecmp($point1['name'], $point2['name']);
  }

}
