<?php

/**
 * @file
 * Contains \Drupal\cdek_api\Element\Select.
 */

namespace Drupal\cdek_api\Element;

use Drupal\Core\Render\Element\CompositeFormElementTrait;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Render\Element\FormElement;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Render\Element;

/**
 * Provides a form element for selecting an order pickup point.
 *
 * @FormElement("cdek_api_select")
 */
class Select extends FormElement {

  use CompositeFormElementTrait;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#input' => TRUE,
      '#process' => [[$class, 'processSelect']],
      '#after_build' => [[$class, 'afterBuildSelect']],
      '#pre_render' => [[$class, 'preRenderCompositeFormElement']],
      '#theme_wrappers' => ['container__cdek_api_select'],
      '#cdek_api_options' => [],
      '#cdek_api_params' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    $value = [];

    if ($input === FALSE) {
      if (isset($element['#default_value'])) {
        /** @var \Drupal\cdek_api\Cdek $cdek */
        $cdek = \Drupal::service('cdek_api');

        // Trying to find the order pickup point.
        if ($point = $cdek->getOrderPickupPoint($element['#default_value'])) {
          $value = [
            'country' => self::isCountryElementDisplayed($element) ? 'countryCode' : NULL,
            'region' => self::isRegionElementDisplayed($element) ? 'regionCode' : NULL,
            'city' => self::isCityElementDisplayed($element) ? 'cityCode' : NULL,
            'point' => 'code',
          ];
          $value = array_filter($value);
          foreach ($value as &$item) {
            $item = $point[$item];
          }
        }
      }
    }
    elseif (is_array($input)) {
      $value = $input;
    }
    return $value;
  }

  /**
   * Processes a form element for selecting an order pickup point.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param array $complete_form
   *   The complete form structure.
   *
   * @return array
   *   The processed element.
   */
  public static function processSelect(&$element, FormStateInterface $form_state, &$complete_form) {
    $value = is_array($element['#value']) ? $element['#value'] : [];
    $params = $element['#cdek_api_params'];

    // Set some default element properties.
    $element['#tree'] = TRUE;

    // Generate a unique wrapper HTML ID.
    $ajax_wrapper_id = Html::getUniqueId('ajax-wrapper');
    $ajax_settings = [
      'callback' => [get_called_class(), 'updateAjaxCallback'],
      'options' => [
        'query' => [
          'element_parents' => implode('/', $element['#array_parents']),
        ],
      ],
      'wrapper' => $ajax_wrapper_id,
      'effect' => 'fade',
    ];

    /** @var \Drupal\cdek_api\Cdek $cdek */
    $cdek = \Drupal::service('cdek_api');
    $value_blockers = [];

    if (self::isCountryElementDisplayed($element)) {
      // Country selection does not depend on other elements.
      $countries = $cdek->getCountries($params) ?: [];
      $country = isset($value['country']) && isset($countries[$value['country']]) ? $value['country'] : '';

      // Element to select the country.
      $element['country'] = [
        '#type' => 'select',
        '#title' => t('Country'),
        '#options' => $countries,
        '#empty_value' => '',
        '#value' => $country,
        '#ajax' => $ajax_settings,
      ];

      // If the country is not yet selected, then include it in the list of
      // value blockers. Otherwise, add the country to the parameters.
      if (empty($value['country'])) {
        $value_blockers[] = 'country';
      }
      else {
        $params['countryid'] = $value['country'];
      }
    }

    if (self::isRegionElementDisplayed($element)) {
      // The region selection may depend on the selected country.
      $regions = empty($value_blockers) ? ($cdek->getRegions($params) ?: []) : [];
      $region = isset($value['region']) && isset($regions[$value['region']]) ? $value['region'] : '';

      // Element to select the region.
      $element['region'] = [
        '#type' => 'select',
        '#title' => t('Region'),
        '#options' => $regions,
        '#empty_value' => '',
        '#value' => $region,
        '#ajax' => $ajax_settings,
      ];

      // If the region is not yet selected, then include it in the list of
      // value blockers. Otherwise, add the region to the parameters.
      if (empty($value['region'])) {
        $value_blockers[] = 'region';
      }
      else {
        $params['regionid'] = $value['region'];
      }
    }

    if (self::isCityElementDisplayed($element)) {
      // The city selection may depend on the selected country and region.
      $cities = empty($value_blockers) ? ($cdek->getCities($params) ?: []) : [];
      $city = isset($value['city']) && isset($cities[$value['city']]) ? $value['city'] : '';

      // Element to select the city.
      $element['city'] = [
        '#type' => 'select',
        '#title' => t('City'),
        '#options' => $cities,
        '#empty_value' => '',
        '#value' => $city,
        '#ajax' => $ajax_settings,
      ];

      // If the city is not yet selected, then include it in the list of
      // value blockers. Otherwise, add the city to the parameters.
      if (empty($value['city'])) {
        $value_blockers[] = 'city';
      }
      else {
        $params['cityid'] = $value['city'];
      }
    }

    // The point selection may depend on the selected country, region and city.
    $points = empty($value_blockers) ? ($cdek->getOrderPickupPoints($params) ?: []) : [];
    $point = isset($value['point']) && isset($points[$value['point']]) ? $value['point'] : '';
    foreach ($points as &$item) {
      $item = $item['name'];
    }

    // Element to select the point.
    $element['point'] = [
      '#type' => 'select',
      '#title' => t('Point'),
      '#options' => $points,
      '#empty_value' => '',
      '#value' => $point,
    ];

    // Prefix and suffix used for Ajax replacement.
    $element['#prefix'] = '<div id="' . $ajax_wrapper_id . '">';
    $element['#suffix'] = '</div>';
    return $element;
  }

  /**
   * Element after-build handler.
   *
   * @param array $element
   *   The element structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   The element structure.
   */
  public static function afterBuildSelect(array $element, FormStateInterface $form_state) {
    // Update element value.
    $element['#value'] = $element['point']['#value'];
    $form_state->setValueForElement($element, $element['#value']);
    return $element;
  }

  /**
   * Ajax callback to update element.
   *
   * @param array $form
   *   The form structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return array
   *   The element to be rendered.
   */
  public static function updateAjaxCallback(array $form, FormStateInterface $form_state, Request $request) {
    // Get element parents from the request.
    $element_parents = explode('/', $request->query->get('element_parents'));
    // Sanitize element parents before using them.
    $element_parents = array_filter($element_parents, [Element::class, 'child']);
    // Retrieve the element to be rendered.
    return NestedArray::getValue($form, $element_parents);
  }

  /**
   * Determines whether the country element is displayed.
   *
   * @param array $element
   *   The element structure.
   *
   * @return bool
   *   TRUE - displayed, FALSE - not displayed.
   */
  protected static function isCountryElementDisplayed(array $element) {
    $options = $element['#cdek_api_options'];
    $params = $element['#cdek_api_params'];
    return empty($options['hide_country']) && !isset($params['cityid']) && !isset($params['regionid']) && !isset($params['countryid']);
  }

  /**
   * Determines whether the region element is displayed.
   *
   * @param array $element
   *   The element structure.
   *
   * @return bool
   *   TRUE - displayed, FALSE - not displayed.
   */
  protected static function isRegionElementDisplayed(array $element) {
    $options = $element['#cdek_api_options'];
    $params = $element['#cdek_api_params'];
    return empty($options['hide_region']) && !isset($params['cityid']) && !isset($params['regionid']);
  }

  /**
   * Determines whether the city element is displayed.
   *
   * @param array $element
   *   The element structure.
   *
   * @return bool
   *   TRUE - displayed, FALSE - not displayed.
   */
  protected static function isCityElementDisplayed(array $element) {
    $options = $element['#cdek_api_options'];
    $params = $element['#cdek_api_params'];
    return empty($options['hide_city']) && !isset($params['cityid']);
  }

}
