<?php

/**
 * @file
 * Contains \Drupal\cdek_api\Form\AdvancedSettingsForm.
 */

namespace Drupal\cdek_api\Form;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Cache\Cache;

/**
 * CDEK API advanced settings form.
 */
class AdvancedSettingsForm extends ConfigFormBase {

  /**
   * The cache object associated with the 'cdek_api' bin.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * Constructs a \Drupal\cdek_api\Form\AdvancedSettingsForm object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache object associated with the 'cdek_api' bin.
   */
  public function __construct(ConfigFactoryInterface $config_factory, CacheBackendInterface $cache) {
    parent::__construct($config_factory);
    $this->cache = $cache;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('cache.cdek_api')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cdek_api_advanced_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'cdek_api.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Get module config.
    $config = $this->config('cdek_api.settings');

    // API request timeout.
    $form['request_timeout'] = [
      '#type' => 'number',
      '#title' => $this->t('API request timeout'),
      '#description' => $this->t('Float representing the maximum number of seconds the API request may take. Use 0 to wait indefinitely.'),
      '#default_value' => $config->get('request_timeout'),
      '#min' => 0,
      '#step' => 0.01,
      '#max' => 30,
      '#required' => TRUE,
    ];

    // API data cache expire.
    $cache_expire = $config->get('cache_expire');
    if ($cache_expire === NULL) {
      $cache_type = 'none';
      $custom_cache_expire = NULL;
    }
    elseif ($cache_expire === Cache::PERMANENT) {
      $cache_type = 'permanent';
      $custom_cache_expire = NULL;
    }
    else {
      $cache_type = 'custom';
      $custom_cache_expire = $cache_expire;
    }
    $form['cache_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('API data cache expire'),
      '#options' => [
        'none' => $this->t('Do not cache'),
        'permanent' => $this->t('Permanently'),
        'custom' => $this->t('Custom'),
      ],
      '#default_value' => $cache_type,
      '#required' => TRUE,
    ];
    $form['custom_cache_expire'] = [
      '#type' => 'number',
      '#title' => $this->t('Custom cache expire'),
      '#description' => $this->t('Cache expire in seconds.'),
      '#default_value' => $custom_cache_expire,
      '#min' => 1,
      '#step' => 1,
      '#states' => [
        'visible' => [':input[name="cache_type"]' => ['value' => 'custom']],
        'required' => [':input[name="cache_type"]' => ['value' => 'custom']],
        'enabled' => [':input[name="cache_type"]' => ['value' => 'custom']],
      ],
      '#element_validate' => [[$this, 'validateCustomCacheExpire']],
    ];

    // Button to clear the CDEK API cache.
    $form['clear_cache']['#type'] = 'container';
    $form['clear_cache']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Clear API cache'),
      '#submit' => [[$this, 'clearCacheSubmit']],
      '#limit_validation_errors' => [],
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * Validate custom_cache_expire element.
   *
   * @param array $element
   *   The element structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function validateCustomCacheExpire(array $element, FormStateInterface $form_state) {
    if ($form_state->getValue('cache_type') == 'custom' && !$form_state->getValue('custom_cache_expire')) {
      $form_state->setError($element, $this->t('@name field is required.', [
        '@name' => $element['#title'],
      ]));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Get module config.
    $config = $this->config('cdek_api.settings');

    // Set config values.
    $config->set('request_timeout', $form_state->getValue('request_timeout'));
    switch ($form_state->getValue('cache_type')) {
      case 'permanent':
        $cache_expire = Cache::PERMANENT;
        break;

      case 'custom':
        $cache_expire = $form_state->getValue('custom_cache_expire');
        break;

      default:
        // Clear the cache if the API data is not cached.
        $this->cache->deleteAll();
        $cache_expire = NULL;
    }
    $config->set('cache_expire', $cache_expire);

    // Save module config.
    $config->save();
    parent::submitForm($form, $form_state);
  }

  /**
   * Submission handler to clear the CDEK API cache.
   */
  public function clearCacheSubmit() {
    $this->cache->deleteAll();
    drupal_set_message($this->t('The cache has been cleared.'));
  }

}
